/* strutil.c
 * String utility routines
 *
 * $Id:  $
 *
 * based on: Wireshark strutil.c & tvbuff.c
 *   Copyright 1998 Gerald Combs
 *
 * Xplico - Internet Traffic Decoder
 * By Gianluca Costa <g.costa@xplico.org>
 * Copyright 2007 Gianluca Costa & Andrea de Franceschi. Web: www.xplico.org
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdlib.h>
#include <string.h>

#include "strutil.h"
#include "istypes.h"


/*
 * Given a pointer into a data buffer, and to the end of the buffer,
 * find the end of the (putative) line at that position in the data
 * buffer.
 * Return a pointer to the EOL character(s) in "*eol".
 */
const char *find_line_end(const char *data, const char *dataend, const char **eol)
{
    const char *lineend;

    lineend = memchr(data, '\n', dataend - data);
    if (lineend == NULL) {
        /*
         * No LF - line is probably continued in next TCP segment.
         */
        lineend = dataend;
        *eol = dataend;
    } else {
        /*
         * Is the LF at the beginning of the line?
         */
        if (lineend > data) {
            /*
             * No - is it preceded by a carriage return?
             * (Perhaps it's supposed to be, but that's not guaranteed....)
             */
            if (*(lineend - 1) == '\r') {
                /*
                 * Yes.  The EOL starts with the CR.
                 */
                *eol = lineend - 1;
            } else {
                /*
                 * No.  The EOL starts with the LF.
                 */
                *eol = lineend;

                /*
                 * I seem to remember that we once saw lines ending with LF-CR
                 * in an HTTP request or response, so check if it's *followed*
                 * by a carriage return.
                 */
                if (lineend < (dataend - 1) && *(lineend + 1) == '\r') {
                    /*
                     * It's <non-LF><LF><CR>; say it ends with the CR.
                     */
                    lineend++;
                }
            }
        } else {
            /*
             * Yes - the EOL starts with the LF.
             */
            *eol = lineend;
        }

        /*
         * Point to the character after the last character.
         */
        lineend++;
    }
    
    return lineend;
}

/*
 * Get the length of the next token in a line, and the beginning of the
 * next token after that (if any).
 * Return 0 if there is no next token.
 */
int get_token_len(const char *linep, const char *lineend, const char **next_token)
{
    const char *tokenp;
    int token_len;

    tokenp = linep;

    /*
     * Search for a blank, a CR or an LF, or the end of the buffer.
     */
    while (linep < lineend && *linep != ' ' && *linep != '\r' && *linep != '\n')
        linep++;
    token_len = linep - tokenp;

    /*
     * Skip trailing blanks.
     */
    while (linep < lineend && *linep == ' ')
        linep++;

    *next_token = linep;

    return token_len;
}




/*
 * Given a tvbuff, an offset into the tvbuff, and a length that starts
 * at that offset (which may be -1 for "all the way to the end of the
 * tvbuff"), find the end of the (putative) line that starts at the
 * specified offset in the tvbuff, going no further than the specified
 * length.
 *
 * However, treat quoted strings inside the buffer specially - don't
 * treat newlines in quoted strings as line terminators.
 *
 * Return the length of the line (not counting the line terminator at
 * the end), or the amount of data remaining in the buffer if we don't
 * find a line terminator.
 *
 * Set "*next_offset" to the offset of the character past the line
 * terminator, or past the end of the buffer if we don't find a line
 * terminator.
 */
int find_line_end_unquoted(char *tvb, int len, int *next_offset)
{
    int cur_offset, char_offset;
    bool is_quoted;
    unsigned char c;
    int eob_offset;
    int linelen;
    char *tmp;

    eob_offset = len;

    cur_offset = 0;
    is_quoted = FALSE;
    for (;;) {
        /*
         * Is this part of the string quoted?
         */
        char_offset = -1;
        if (is_quoted) {
            /*
             * Yes - look only for the terminating quote.
             */
            tmp = memchr(tvb+cur_offset, '"', len - cur_offset);
            if (tmp != NULL) {
                char_offset = tmp - tvb;
            }
            
        } else {
            /*
             * Look either for a CR, an LF, or a '"'.
             */
            tmp =  memchr(tvb+cur_offset, '\r', len - cur_offset);
            if (tmp == NULL) {
                tmp =  memchr(tvb+cur_offset, '\n', len - cur_offset);
                if (tmp == NULL) {
                    tmp =  memchr(tvb+cur_offset, '"', len - cur_offset);
                }
            }
            if (tmp != NULL) {
                char_offset = tmp - tvb;
            }
        }
        if (char_offset == -1) {
            /*
             * Not found - line is presumably continued in
             * next packet.
             * We pretend the line runs to the end of the tvbuff.
             */
            linelen = eob_offset;
            *next_offset = eob_offset;
            break;
        }

        if (is_quoted) {
            /*
             * We're processing a quoted string.
             * We only looked for ", so we know it's a ";
             * as we're processing a quoted string, it's a
             * closing quote.
             */
            is_quoted = FALSE;
        } else {
            /*
             * OK, what is it?
             */
            c = tvb[char_offset];
            if (c == '"') {
                /*
                 * Un-quoted "; it begins a quoted
                 * string.
                 */
                is_quoted = TRUE;
            } else {
                /*
                 * It's a CR or LF; we've found a line
                 * terminator.
                 *
                 * Find the number of bytes between the
                 * starting offset and the CR or LF.
                 */
                linelen = char_offset;

                /*
                 * Is it a CR?
                 */
                if (c == '\r') {
                    /*
                     * Yes; is it followed by an LF?
                     */
                    if (char_offset + 1 < eob_offset &&
                        tvb[char_offset + 1]
                        == '\n') {
                        /*
                         * Yes; skip over the CR.
                         */
                        char_offset++;
                    }
                }

                /*
                 * Return the offset of the character after
                 * the last character in the line, skipping
                 * over the last character in the line
                 * terminator, and quit.
                 */
                *next_offset = char_offset + 1;
                break;
            }
        }

        /*
         * Step past the character we found.
         */
        cur_offset = char_offset + 1;
        if (cur_offset >= eob_offset) {
            /*
             * The character we found was the last character
             * in the tvbuff - line is presumably continued in
             * next packet.
             * We pretend the line runs to the end of the tvbuff.
             */
            linelen = eob_offset;
            *next_offset = eob_offset;
            break;
        }
    }

    return linelen;
}


int find_chr(const char *line, int len, char c)
{
    int i;
    
    i = 0;
    
    while (i<len && line[i] != c)
        i++;
    if (i == len)
        return -1;

    return i;
}


void strutil_link(void)
{
    
}
